

from . import testing


class CrossSheetTest(testing.FunctionalTestCase):

    filename = "cross_sheet.xlsx"

    def test_reference(self):
        """
            Tests if a reference to a cell on another sheet (Sheet1),
            which refers to a cell on it's own sheet (Sheet2) is resolved
            properly as being on Sheet2

            Also validates if after that passing in only "A1" refers back to
            Sheet1
        """
        # This tests going back and forth between original Sheet1 and Sheet2
        # multiple times in different orders
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!D2'),
            24
        )

        for coord, cell in self.model.cells.items():
            if cell.formula is None or cell.value is None:
                continue
            expected = cell.value
            cell.value = None
            got = self.evaluator.evaluate(coord)
            self.assertEqual(
                got,
                expected,
                msg=f"{coord} got: {got} expected: {expected}"
            )

        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A5'),
            4
        )
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!C5'),
            28
        )

        self.assertEqual(
            self.evaluator.evaluate('Sheet2!C1'),
            12
        )



from .. import testing


class COSHTest(testing.FunctionalTestCase):
    filename = "COSH.xlsx"

    def test_evaluation_A2(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A2')
        value = self.evaluator.evaluate('Sheet1!A2')
        self.assertAlmostEqual(excel_value, value, 8)

    def test_evaluation_A3(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A3')
        value = self.evaluator.evaluate('Sheet1!A3')
        self.assertAlmostEqual(excel_value, value)



from .. import testing


class SumTest(testing.FunctionalTestCase):
    filename = "BASES.xlsx"

    def test_evalMatrix(self):
        for col in 'BCDEFGHIJKLM':
            for row in range(4, 20):
                addr = f'Sheet1!{col}{row}'
                excel_value = self.evaluator.get_cell_value(addr)
                value = self.evaluator.evaluate(addr)
                self.assertEqual(excel_value, value, addr)

    def test_converstionWithReferences(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!R14')
        value = self.evaluator.evaluate('Sheet1!R14')
        self.assertEqual(excel_value, value)



from .. import testing


class CountIfTest(testing.FunctionalTestCase):
    filename = "IF.xlsx"

    def test_evaluation_ABCDE_1(self):
        for col in "ABCDE":
            cell = f'Sheet1!{col}1'
            excel_value = self.evaluator.get_cell_value(cell)
            value = self.evaluator.evaluate(cell)
            self.assertEqual(excel_value, value)



from .. import testing


class EDateTest(testing.FunctionalTestCase):
    filename = "EOMONTH.xlsx"

    def test_counta_evaluation_A1(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A1')
        value = self.evaluator.evaluate('Sheet1!A1')
        self.assertEqual(excel_value, value)

    def test_evaluation_B1(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!B1')
        value = self.evaluator.evaluate('Sheet1!B1')
        self.assertEqual(excel_value, value)



from typing import Literal, Optional, Union

from . import xl, func_xltypes
from .func_xltypes import UNUSED, Unused
from .xlerrors import NumExcelError, ValueExcelError

dec = "dec"
Base = Literal[bin, dec, oct, hex]

PERMITTED_DIGITS = {
    bin: set("01"),
    oct: set("01234567"),
    hex: set("0123456789ABCDEFabcdef"),
}


BIT_WIDTHS = {bin: 10, oct: 30, hex: 40}

BASE_NUMBERS = {bin: 2, oct: 8, hex: 16}


BOUNDS = {
    frozenset([bin, oct]): 2 ** 9,
    frozenset([bin, dec]): 2 ** 9,
    frozenset([bin, hex]): 2 ** 9,
    frozenset([oct, dec]): 2 ** 29,
    frozenset([oct, hex]): 2 ** 29,
    frozenset([dec, hex]): 2 ** 39,
}


def handle_places(
        places: Union[Unused, func_xltypes.XlAnything]
) -> Optional[int]:
    if places is UNUSED:
        return None

    if isinstance(places, func_xltypes.Boolean):
        raise ValueExcelError('The `places` argument cannot be a boolean.')

    places = int(places)
    if not (1 <= places <= 10):
        raise NumExcelError('The number of places must be between 1 and 10.')

    return places


def handle_number(number: func_xltypes.XlAnything, origin) -> Union[int, str]:
    if isinstance(number, func_xltypes.Boolean):
        raise ValueExcelError('The number cannot be a boolean.')

    if origin == dec:
        return int(number)

    if isinstance(number, func_xltypes.Blank):
        as_str = "0"

    elif isinstance(number, func_xltypes.Number):
        if number.is_decimal and not number.value.is_integer():
            raise NumExcelError('Number is not an integer.')

        as_str = str(int(number))

    elif isinstance(number, func_xltypes.Text):
        as_str = str(number) if number else "0"

    if len(as_str) > 10:
        raise NumExcelError()

    if set(as_str) - PERMITTED_DIGITS[origin]:
        raise NumExcelError('There are invalid characters in the number.')

    return as_str


def pad_zeroes(string, was_negative, places):
    if places is None:
        return string

    desired_length = len(string) if was_negative else places
    if desired_length < len(string):
        raise NumExcelError(
            'The resulting string is longer than the desired length.')

    return string.zfill(desired_length)


def conversion(number, origin, destination, places):
    if origin == dec:
        # if origin base is dec we do not need to convert from it
        value = number
    else:
        # otherwise, convert to int in the appropriate base.
        as_int = int(number, BASE_NUMBERS[origin])

        # magic:
        mask = 1 << BIT_WIDTHS[origin] - 1
        value = (as_int & ~mask) - (as_int & mask)

    # This is another error-check, but it had to be delayed to here rather
    # than in handle_number because we had to convert it from the origin base
    # first.
    bound = BOUNDS[frozenset([origin, destination])]
    if not (-bound <= value < bound):
        raise NumExcelError('The input number is out of bounds')

    # if the destination base is dec, we are done
    if destination == dec:
        return value

    # magic (handles the 2s-complement-like wrapping behavior):
    was_negative = value < 0
    if was_negative:
        value += 1 << BIT_WIDTHS[destination]

    # Otherwise convert to appropriate string representation (return value is
    # (str, bool)).
    result = destination(value)[2:].upper()
    return pad_zeroes(result, was_negative, places)


def convert_bases(number, origin, destination, places=None):
    # If places is None, means it was not passed to this function, i.e. the
    # corresponding Excel function does not take it as an argument (the
    # ___2DEC functions). If it was passed in, we do error checking via the
    # handle_places function.
    if places is not None:
        places = handle_places(places)

    # Next, we do error checking on the number argument. This function does
    # not do any base-conversion logic, it simply raises exceptions if it
    # encounters invalid inputs.  If the origin base is dec, the result is an
    # integer, otherwise it is a string.
    number = handle_number(number, origin)

    # Now we actually convert from one base to another. The result of this can
    # also be either and integer or a string depending on the destination
    # base. If it is a string, it may also be zero-padded depending on the
    # places argument.
    return conversion(number, origin, destination, places)


@xl.register()
@xl.validate_args
def DEC2BIN(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, dec, bin, places)


@xl.register()
@xl.validate_args
def DEC2OCT(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, dec, oct, places)


@xl.register()
@xl.validate_args
def DEC2HEX(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, dec, hex, places)


# the ___2DEC functions give a number not a string, and they do not take a
# `places` parameter
@xl.register()
@xl.validate_args
def BIN2DEC(
        number: func_xltypes.XlAnything
) -> func_xltypes.XlNumber:
    return convert_bases(number, bin, dec)


@xl.register()
@xl.validate_args
def BIN2OCT(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, bin, oct, places)


@xl.register()
@xl.validate_args
def BIN2HEX(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, bin, hex, places)


@xl.register()
@xl.validate_args
def OCT2DEC(
        number: func_xltypes.XlAnything
) -> func_xltypes.XlNumber:
    return convert_bases(number, oct, dec)


@xl.register()
@xl.validate_args
def OCT2BIN(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, oct, bin, places)


@xl.register()
@xl.validate_args
def OCT2HEX(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, oct, hex, places)


@xl.register()
@xl.validate_args
def HEX2DEC(
        number: func_xltypes.XlAnything
) -> func_xltypes.XlNumber:
    return convert_bases(number, hex, dec)


@xl.register()
@xl.validate_args
def HEX2BIN(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, hex, bin, places)


@xl.register()
@xl.validate_args
def HEX2OCT(
        number: func_xltypes.XlAnything,
        places: func_xltypes.XlAnything = UNUSED
) -> func_xltypes.XlText:
    return convert_bases(number, hex, oct, places)



import unittest

from xlcalculator.xlfunctions import xlerrors, func_xltypes, engineering


class EngineeringModuleTest(unittest.TestCase):

    def test_DEC2BIN(self):
        self.assertEqual(
            engineering.DEC2BIN(10), func_xltypes.Text('1010'))

    def test_DEC2BIN_withPlaces(self):
        self.assertEqual(
            engineering.DEC2BIN(10, 5), func_xltypes.Text('01010'))

    def test_DEC2BIN_withBlank(self):
        self.assertEqual(
            engineering.DEC2BIN(func_xltypes.Blank()), func_xltypes.Text('0'))

    def test_DEC2BIN_withNegativeNumbers(self):
        self.assertEqual(engineering.DEC2BIN(-1), '1111111111')
        self.assertEqual(engineering.DEC2BIN(-2), '1111111110')
        self.assertEqual(engineering.DEC2BIN(-4.2), '1111111100')
        self.assertEqual(engineering.DEC2BIN(-4.8), '1111111100')

    def test_DEC2BIN_withBooleanNumber(self):
        self.assertIsInstance(
            engineering.DEC2BIN(True), xlerrors.ValueExcelError)

    def test_DEC2BIN_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.DEC2BIN(10, 20), xlerrors.NumExcelError)

    def test_DEC2BIN_withInsufficientPlaces(self):
        self.assertIsInstance(
            engineering.DEC2BIN(10, 3), xlerrors.NumExcelError)

    def test_DEC2BIN_withBooleanPlaces(self):
        self.assertIsInstance(
            engineering.DEC2BIN(10, True), xlerrors.ValueExcelError)

    def test_DEC2BIN_numberTooLarge(self):
        self.assertIsInstance(
            engineering.DEC2BIN(512), xlerrors.NumExcelError)
        self.assertIsInstance(
            engineering.DEC2BIN(-513), xlerrors.NumExcelError)

    def test_DEC2OCT(self):
        self.assertEqual(
            engineering.DEC2OCT(8), func_xltypes.Text('10'))
        self.assertEqual(
            engineering.DEC2OCT(-1), func_xltypes.Text('7777777777'))
        self.assertEqual(
            engineering.DEC2OCT(-2), func_xltypes.Text('7777777776'))

    def test_DEC2OCT_checkBounds(self):
        self.assertEqual(
            engineering.DEC2OCT(2**29 - 1), func_xltypes.Text('3777777777'))
        self.assertEqual(
            engineering.DEC2OCT(-(2**29)), func_xltypes.Text('4000000000'))
        self.assertIsInstance(
            engineering.DEC2OCT(2**29), xlerrors.NumExcelError)
        self.assertIsInstance(
            engineering.DEC2OCT(-(2**29) - 1), xlerrors.NumExcelError)

    def test_DEC2OCT_withPlaces(self):
        self.assertEqual(
            engineering.DEC2OCT(8, 5), func_xltypes.Text('00010'))
        self.assertEqual(
            engineering.DEC2OCT(-1, 1), func_xltypes.Text('7777777777'))

    def test_DEC2OCT_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.DEC2OCT(8, 1), xlerrors.NumExcelError)

    def test_DEC2HEX(self):
        self.assertEqual(
            engineering.DEC2HEX(255), func_xltypes.Text('FF'))
        self.assertEqual(
            engineering.DEC2HEX(-1), func_xltypes.Text('FFFFFFFFFF'))
        self.assertEqual(
            engineering.DEC2HEX(-2), func_xltypes.Text('FFFFFFFFFE'))

    def test_DEC2HEX_checkBounds(self):
        self.assertEqual(
            engineering.DEC2HEX(2**39 - 1), func_xltypes.Text('7FFFFFFFFF'))
        self.assertEqual(
            engineering.DEC2HEX(-(2**39)), func_xltypes.Text('8000000000'))
        self.assertIsInstance(
            engineering.DEC2HEX(2**39), xlerrors.NumExcelError)
        self.assertIsInstance(
            engineering.DEC2HEX(-(2**39) - 1), xlerrors.NumExcelError)

    def test_DEC2HEX_withPlaces(self):
        self.assertEqual(
            engineering.DEC2HEX(16, 5), func_xltypes.Text('00010'))
        self.assertEqual(
            engineering.DEC2HEX(-1, 1), func_xltypes.Text('FFFFFFFFFF'))

    def test_DEC2HEX_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.DEC2HEX(16, 1), xlerrors.NumExcelError)

    def test_BIN2DEC(self):
        self.assertEqual(
            engineering.BIN2DEC('1010'), 10)
        self.assertEqual(
            engineering.BIN2DEC(1010), 10)

    def test_BIN2DEC_withBlank(self):
        self.assertEqual(
            engineering.BIN2DEC(func_xltypes.Blank()), func_xltypes.Number(0))

    def test_BIN2DEC_withNonInteger(self):
        self.assertIsInstance(
            engineering.BIN2DEC(func_xltypes.Number(1.1)),
            xlerrors.NumExcelError)

    def test_BIN2DEC_numberTooLarge(self):
        self.assertIsInstance(
            engineering.BIN2DEC(10**20), xlerrors.NumExcelError)

    def test_BIN2DEC_withInvalidNumber(self):
        self.assertIsInstance(
            engineering.BIN2DEC('bad'), xlerrors.NumExcelError)

    def test_BIN2OCT(self):
        self.assertEqual(
            engineering.BIN2OCT(10000000), func_xltypes.Text('200'))
        self.assertEqual(
            engineering.BIN2OCT(1000000000), func_xltypes.Text('7777777000'))

    def test_BIN2OCT_checkBounds(self):
        self.assertEqual(
            engineering.BIN2OCT(1111111111), func_xltypes.Text('7777777777'))
        self.assertIsInstance(
            engineering.BIN2OCT(2), xlerrors.NumExcelError)
        self.assertIsInstance(
            engineering.BIN2OCT(10000000000), xlerrors.NumExcelError)

    def test_BIN2OCT_withPlaces(self):
        self.assertEqual(
            engineering.BIN2OCT(1000, 5), func_xltypes.Text('00010'))
        self.assertEqual(
            engineering.BIN2OCT(1000000000, 1),
            func_xltypes.Text('7777777000'))

    def test_BIN2OCT_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.BIN2OCT(1000, 1), xlerrors.NumExcelError)

    def test_BIN2HEX(self):
        self.assertEqual(
            engineering.BIN2HEX(10), func_xltypes.Text('2'))
        self.assertEqual(
            engineering.BIN2HEX(10000000), func_xltypes.Text('80'))

    def test_BIN2HEX_checkBounds(self):
        self.assertEqual(
            engineering.BIN2HEX(1111111111), func_xltypes.Text('FFFFFFFFFF'))
        self.assertIsInstance(
            engineering.BIN2HEX(11000000000), xlerrors.NumExcelError)
        self.assertIsInstance(
            engineering.BIN2HEX(10000000000), xlerrors.NumExcelError)

    def test_BIN2HEX_withPlaces(self):
        self.assertEqual(
            engineering.BIN2HEX(1000, 5), func_xltypes.Text('00008'))
        self.assertEqual(
            engineering.BIN2HEX(1000000000, 1),
            func_xltypes.Text('FFFFFFFE00'))

    def test_BIN2HEX_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.BIN2HEX(10000, 1), xlerrors.NumExcelError)

    def test_OCT2DEC(self):
        self.assertEqual(
            engineering.OCT2DEC(10), 8)
        self.assertEqual(
            engineering.OCT2DEC(7777777777), -1)

    def test_OCT2DEC_checkBounds(self):
        self.assertEqual(
            engineering.OCT2DEC(3777777777), 536870911)
        self.assertEqual(
            engineering.OCT2DEC(4000000000), -536870912)
        self.assertIsInstance(
            engineering.OCT2DEC(10000000000), xlerrors.NumExcelError)

    def test_OCT2BIN(self):
        self.assertEqual(
            engineering.OCT2BIN(2), func_xltypes.Text('10'))
        self.assertEqual(
            engineering.OCT2BIN(777), func_xltypes.Text('111111111'))

    def test_OCT2BIN_checkBounds(self):
        self.assertEqual(
            engineering.OCT2BIN(7777777000), func_xltypes.Text('1000000000'))
        self.assertIsInstance(
            engineering.OCT2BIN(1000), xlerrors.NumExcelError)

    def test_OCT2BIN_withPlaces(self):
        self.assertEqual(
            engineering.OCT2BIN(2, 5), func_xltypes.Text('00010'))
        self.assertEqual(
            engineering.OCT2BIN(7777777000, 1),
            func_xltypes.Text('1000000000'))

    def test_OCT2BIN_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.OCT2BIN(2, 1), xlerrors.NumExcelError)

    def test_OCT2HEX(self):
        self.assertEqual(
            engineering.OCT2HEX(10), func_xltypes.Text('8'))
        self.assertEqual(
            engineering.OCT2HEX(12), func_xltypes.Text('A'))

    def test_OCT2HEX_checkBounds(self):
        self.assertEqual(
            engineering.OCT2HEX(7777777777), func_xltypes.Text('FFFFFFFFFF'))
        self.assertIsInstance(
            engineering.OCT2HEX(10000000000), xlerrors.NumExcelError)

    def test_OCT2HEX_withPlaces(self):
        self.assertEqual(
            engineering.OCT2HEX(20, 5), func_xltypes.Text('00010'))
        self.assertEqual(
            engineering.OCT2HEX(4000000000, 1),
            func_xltypes.Text('FFE0000000'))

    def test_OCT2HEX_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.OCT2HEX(20, 1), xlerrors.NumExcelError)

    def test_HEX2DEC(self):
        self.assertEqual(
            engineering.HEX2DEC(10), 16)
        self.assertEqual(
            engineering.HEX2DEC('fF'), 255)

    def test_HEX2DEC_checkBounds(self):
        self.assertEqual(
            engineering.HEX2DEC('FFFFFFFFFF'), -1)
        self.assertEqual(
            engineering.HEX2DEC(8000000000), -549755813888)
        self.assertIsInstance(
            engineering.HEX2DEC(10000000000), xlerrors.NumExcelError)

    def test_HEX2BIN(self):
        self.assertEqual(
            engineering.HEX2BIN(2), func_xltypes.Text('10'))
        self.assertEqual(
            engineering.HEX2BIN('1E0'), func_xltypes.Text('111100000'))

    def test_HEX2BIN_checkBounds(self):
        self.assertEqual(
            engineering.HEX2BIN('1FF'), func_xltypes.Text('111111111'))
        self.assertIsInstance(
            engineering.HEX2BIN('FFFFFFFDFF'), xlerrors.NumExcelError)

    def test_HEX2BIN_withPlaces(self):
        self.assertEqual(
            engineering.HEX2BIN(2, 5), func_xltypes.Text('00010'))
        self.assertEqual(
            engineering.HEX2BIN('FFFFFFFE00', 1),
            func_xltypes.Text('1000000000'))

    def test_HEX2BIN_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.HEX2BIN(20, 1), xlerrors.NumExcelError)

    def test_HEX2OCT(self):
        self.assertEqual(
            engineering.HEX2OCT(8), func_xltypes.Text('10'))
        self.assertEqual(
            engineering.HEX2OCT('1E3'), func_xltypes.Text('743'))

    def test_HEX2OCT_checkBounds(self):
        self.assertEqual(
            engineering.HEX2OCT('1FFFFFFF'), func_xltypes.Text('3777777777'))
        self.assertEqual(
            engineering.HEX2OCT('FFE0000000'), func_xltypes.Text('4000000000'))
        self.assertIsInstance(
            engineering.HEX2OCT('FFDFFFFFFF'), xlerrors.NumExcelError)

    def test_HEX2OCT_withPlaces(self):
        self.assertEqual(
            engineering.HEX2OCT(8, 5), func_xltypes.Text('00010'))
        self.assertEqual(
            engineering.HEX2OCT('FFFFFFFE00', 1),
            func_xltypes.Text('7777777000'))

    def test_HEX2OCT_withTooManyPlaces(self):
        self.assertIsInstance(
            engineering.HEX2OCT(8, 1), xlerrors.NumExcelError)
